"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const anchor_1 = require("@coral-xyz/anchor");
const web3_js_1 = require("@solana/web3.js");
const fs_1 = __importDefault(require("fs"));
const perp_1 = require("../../src/accounts/perp");
const client_1 = require("../../src/client");
const constants_1 = require("../../src/constants");
const I80F48_1 = require("../../src/numbers/I80F48");
const utils_1 = require("../../src/utils");
// For easy switching between mainnet and devnet, default is mainnet
const CLUSTER = process.env.CLUSTER_OVERRIDE || 'mainnet-beta';
const CLUSTER_URL = process.env.CLUSTER_URL_OVERRIDE || process.env.MB_CLUSTER_URL;
const USER_KEYPAIR = process.env.USER_KEYPAIR_OVERRIDE || process.env.MB_PAYER_KEYPAIR;
const MANGO_ACCOUNT_PK = process.env.MANGO_ACCOUNT_PK || '';
async function settlePnl(mangoAccount, perpMarket, client, group) {
    if (!mangoAccount.perpPositionExistsForMarket(perpMarket)) {
        return;
    }
    const pp = mangoAccount
        .perpActive()
        .find((pp) => pp.marketIndex === perpMarket.perpMarketIndex);
    const pnl = pp.getUnsettledPnl(perpMarket);
    console.log(`Avg entry price - ${pp.getAverageEntryPriceUi(perpMarket)}, Breakeven price - ${pp.getBreakEvenPriceUi(perpMarket)}`);
    let profitableAccount, unprofitableAccount;
    if (pnl.abs().gt((0, utils_1.toNativeI80F48)(1, 6))) {
        console.log(`- Settling pnl ${(0, utils_1.toUiDecimalsForQuote)(pnl)} ...`);
    }
    else {
        console.log(`- Skipping Settling pnl ${(0, utils_1.toUiDecimalsForQuote)(pnl)}, too small`);
        return;
    }
    if (pnl.gt((0, I80F48_1.ZERO_I80F48)())) {
        console.log(`- Settling profit pnl...`);
        profitableAccount = mangoAccount;
        const candidates = await perpMarket.getSettlePnlCandidates(client, group, undefined, 'negative');
        if (candidates.length === 0) {
            return;
        }
        unprofitableAccount = candidates[0].account;
        const sig = await client.perpSettlePnl(group, profitableAccount, unprofitableAccount, mangoAccount, perpMarket.perpMarketIndex);
        console.log(`- Settled pnl, sig https://explorer.solana.com/tx/${sig}?cluster=${CLUSTER == 'devnet' ? 'devnet' : ''}`);
    }
    else if (pnl.lt((0, I80F48_1.ZERO_I80F48)())) {
        unprofitableAccount = mangoAccount;
        const candidates = await perpMarket.getSettlePnlCandidates(client, group, undefined, 'positive');
        if (candidates.length === 0) {
            return;
        }
        profitableAccount = candidates[0].account;
        console.log(`- Settling loss pnl...`);
        let sig = await client.perpSettlePnl(group, profitableAccount, unprofitableAccount, mangoAccount, perpMarket.perpMarketIndex);
        console.log(`- Settled pnl, sig https://explorer.solana.com/tx/${sig}?cluster=${CLUSTER == 'devnet' ? 'devnet' : ''}`);
    }
}
async function takeOrder(client, group, mangoAccount, perpMarket, side) {
    await mangoAccount.reload(client);
    const size = Math.random() * 0.001;
    const price = side === perp_1.PerpOrderSide.bid
        ? perpMarket.uiPrice * 1.01
        : perpMarket.uiPrice * 0.99;
    console.log(`- ${perpMarket.name} taking with a ${side === perp_1.PerpOrderSide.bid ? 'bid' : 'ask'} at  price ${price.toFixed(4)} and size ${size.toFixed(6)}`);
    const oldPosition = mangoAccount.getPerpPosition(perpMarket.perpMarketIndex);
    if (oldPosition) {
        console.log(`-- before base: ${perpMarket.baseLotsToUi(oldPosition.basePositionLots)}, quote: ${(0, utils_1.toUiDecimalsForQuote)(oldPosition.quotePositionNative)}`);
    }
    await client.perpPlaceOrder(group, mangoAccount, perpMarket.perpMarketIndex, side, price, size, undefined, Date.now(), perp_1.PerpOrderType.market, false, 0, 10);
    // Sleep to see change, alternatively we could reload account with processed commitmment
    await new Promise((r) => setTimeout(r, 5000));
    await mangoAccount.reload(client);
    const newPosition = mangoAccount.getPerpPosition(perpMarket.perpMarketIndex);
    if (newPosition) {
        console.log(`-- after base: ${perpMarket.baseLotsToUi(newPosition.basePositionLots)}, quote: ${(0, utils_1.toUiDecimalsForQuote)(newPosition.quotePositionNative)}`);
    }
}
async function main() {
    const options = anchor_1.AnchorProvider.defaultOptions();
    const connection = new web3_js_1.Connection(CLUSTER_URL, options);
    const user = web3_js_1.Keypair.fromSecretKey(Buffer.from(JSON.parse(fs_1.default.readFileSync(USER_KEYPAIR, 'utf-8'))));
    const userWallet = new anchor_1.Wallet(user);
    const userProvider = new anchor_1.AnchorProvider(connection, userWallet, options);
    const client = await client_1.MangoClient.connect(userProvider, CLUSTER, constants_1.MANGO_V4_ID[CLUSTER], {
        idsSource: 'get-program-accounts',
    });
    // Load mango account
    let mangoAccount = await client.getMangoAccount(new web3_js_1.PublicKey(MANGO_ACCOUNT_PK));
    await mangoAccount.reload(client);
    // Load group
    const group = await client.getGroup(mangoAccount.group);
    await group.reloadAll(client);
    // Take on OB
    const perpMarket = group.getPerpMarketByName('BTC-PERP');
    while (true) {
        await group.reloadAll(client);
        // Settle pnl
        await settlePnl(mangoAccount, perpMarket, client, group);
        await takeOrder(client, group, mangoAccount, perpMarket, perp_1.PerpOrderSide.bid);
        await takeOrder(client, group, mangoAccount, perpMarket, perp_1.PerpOrderSide.ask);
    }
}
main();
