"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const listingTools_1 = require("@blockworks-foundation/mango-v4-settings/lib/helpers/listingTools");
const anchor_1 = require("@coral-xyz/anchor");
const spl_governance_1 = require("@solana/spl-governance");
const web3_js_1 = require("@solana/web3.js");
const fs_1 = __importDefault(require("fs"));
const client_1 = require("../src/client");
const constants_1 = require("../src/constants");
const constants_2 = require("./governanceInstructions/constants");
const createProposal_1 = require("./governanceInstructions/createProposal");
const voteStakeRegistryClient_1 = require("./governanceInstructions/voteStakeRegistryClient");
const { MB_CLUSTER_URL, PROPOSAL_TITLE, PROPOSAL_LINK, VSR_DELEGATE_KEYPAIR, VSR_DELEGATE_FROM_PK, DRY_RUN, } = process.env;
async function buildClient() {
    return await client_1.MangoClient.connectDefault(MB_CLUSTER_URL);
}
async function setupWallet() {
    const clientKeypair = web3_js_1.Keypair.fromSecretKey(Buffer.from(JSON.parse(fs_1.default.readFileSync(VSR_DELEGATE_KEYPAIR, 'utf-8'))));
    const clientWallet = new anchor_1.Wallet(clientKeypair);
    return clientWallet;
}
async function setupVsr(connection, clientWallet) {
    const options = anchor_1.AnchorProvider.defaultOptions();
    const provider = new anchor_1.AnchorProvider(connection, clientWallet, options);
    const vsrClient = await voteStakeRegistryClient_1.VsrClient.connect(provider, voteStakeRegistryClient_1.DEFAULT_VSR_ID);
    return vsrClient;
}
async function updateSpotMarkets() {
    const [client, wallet] = await Promise.all([buildClient(), setupWallet()]);
    const vsrClient = await setupVsr(client.connection, wallet);
    const group = await client.getGroup(constants_1.MANGO_V4_MAIN_GROUP);
    const instructions = [];
    Array.from(group.banksMapByTokenIndex.values())
        .map((banks) => banks[0])
        .sort((a, b) => a.name.localeCompare(b.name))
        .forEach(async (bank) => {
        let change = false;
        const tier = Object.values(listingTools_1.LISTING_PRESETS).find((x) => x.initLiabWeight.toFixed(1) === '1.8'
            ? x.initLiabWeight.toFixed(1) ===
                bank?.initLiabWeight.toNumber().toFixed(1) &&
                x.reduceOnly === bank.reduceOnly
            : x.initLiabWeight.toFixed(1) ===
                bank?.initLiabWeight.toNumber().toFixed(1));
        let reduceOnly = null;
        let forceClose = null;
        const name = null;
        const oraclePriceBand = null;
        let markets = [];
        if (bank.reduceOnly == 1 && bank.forceClose && bank.forceWithdraw) {
            markets = Array.from(group.serum3MarketsMapByMarketIndex.values()).filter((m) => m.baseTokenIndex == bank.tokenIndex || m.quoteTokenIndex == 1);
            change = true;
            reduceOnly = true;
            forceClose = true;
            console.log(`${bank.name} ${markets.map((m) => m.name).join(',')}`);
        }
        for (const market of markets) {
            const ix = await client.program.methods
                .serum3EditMarket(reduceOnly, forceClose, name, oraclePriceBand)
                .accounts({
                group: group.publicKey,
                admin: group.admin,
                market: market.publicKey,
            })
                .instruction();
            const tx = new web3_js_1.Transaction({ feePayer: wallet.publicKey }).add(ix);
            const simulated = await client.connection.simulateTransaction(tx);
            if (simulated.value.err) {
                console.log('sim error', simulated.value.logs);
                throw simulated.value.logs;
            }
            if (change) {
                instructions.push(ix);
            }
        }
    });
    const tokenOwnerRecordPk = await (0, spl_governance_1.getTokenOwnerRecordAddress)(constants_2.MANGO_GOVERNANCE_PROGRAM, constants_2.MANGO_REALM_PK, constants_2.MANGO_MINT, new web3_js_1.PublicKey(VSR_DELEGATE_FROM_PK));
    const [tokenOwnerRecord, proposals] = await Promise.all([
        (0, spl_governance_1.getTokenOwnerRecord)(client.connection, tokenOwnerRecordPk),
        (0, spl_governance_1.getAllProposals)(client.connection, constants_2.MANGO_GOVERNANCE_PROGRAM, constants_2.MANGO_REALM_PK),
    ]);
    const walletSigner = wallet;
    if (!DRY_RUN) {
        const proposalAddress = await (0, createProposal_1.createProposal)(client.connection, walletSigner, constants_2.MANGO_DAO_WALLET_GOVERNANCE, tokenOwnerRecord, PROPOSAL_TITLE ? PROPOSAL_TITLE : 'Update spot markets in mango-v4', PROPOSAL_LINK ?? '', Object.values(proposals).length, instructions, vsrClient, false);
        console.log(proposalAddress.toBase58());
    }
}
async function main() {
    try {
        await updateSpotMarkets();
    }
    catch (error) {
        console.log(error);
    }
}
try {
    main();
}
catch (error) {
    console.log(error);
}
