"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const anchor_1 = require("@coral-xyz/anchor");
const spl_governance_1 = require("@solana/spl-governance");
const spl_token_1 = require("@solana/spl-token");
const web3_js_1 = require("@solana/web3.js");
const fs_1 = __importDefault(require("fs"));
const client_1 = require("../src/client");
const constants_1 = require("../src/constants");
const I80F48_1 = require("../src/numbers/I80F48");
const utils_1 = require("../src/utils");
const rpc_1 = require("../src/utils/rpc");
const constants_2 = require("./governanceInstructions/constants");
const createProposal_1 = require("./governanceInstructions/createProposal");
const voteStakeRegistryClient_1 = require("./governanceInstructions/voteStakeRegistryClient");
const { MB_CLUSTER_URL, PROPOSAL_TITLE, PROPOSAL_LINK, VSR_DELEGATE_KEYPAIR, VSR_DELEGATE_FROM_PK, DRY_RUN, } = process.env;
async function buildClient() {
    return await client_1.MangoClient.connectDefault(MB_CLUSTER_URL);
}
async function setupWallet() {
    const clientKeypair = web3_js_1.Keypair.fromSecretKey(Buffer.from(JSON.parse(fs_1.default.readFileSync(VSR_DELEGATE_KEYPAIR, 'utf-8'))));
    const clientWallet = new anchor_1.Wallet(clientKeypair);
    return clientWallet;
}
async function setupVsr(connection, clientWallet) {
    const options = anchor_1.AnchorProvider.defaultOptions();
    const provider = new anchor_1.AnchorProvider(connection, clientWallet, options);
    const vsrClient = await voteStakeRegistryClient_1.VsrClient.connect(provider, voteStakeRegistryClient_1.DEFAULT_VSR_ID);
    return vsrClient;
}
async function withdrawFeesToAdmin() {
    const [client, wallet] = await Promise.all([buildClient(), setupWallet()]);
    const vsrClient = await setupVsr(client.connection, wallet);
    // create wsol ata
    const ix = await (0, utils_1.createAssociatedTokenAccountIdempotentInstruction)(wallet.publicKey, new web3_js_1.PublicKey('8SSLjXBEVk9nesbhi9UMCA32uijbVBUqWoKPPQPTekzt'), new web3_js_1.PublicKey('So11111111111111111111111111111111111111112'));
    const res = await (0, rpc_1.sendTransaction)(vsrClient.program.provider, [ix], [], {});
    console.log(`${res.signature}`);
    const group = await client.getGroup(constants_1.MANGO_V4_MAIN_GROUP);
    const instructions = [];
    Array.from(group.banksMapByTokenIndex.values())
        .map((banks) => banks[0])
        .sort((a, b) => a.name.localeCompare(b.name))
        .forEach(async (bank) => {
        const tokenAccount = await (0, spl_token_1.getAssociatedTokenAddress)(bank.mint, new web3_js_1.PublicKey('8SSLjXBEVk9nesbhi9UMCA32uijbVBUqWoKPPQPTekzt'));
        const fees = bank.collectedFeesNative.sub(I80F48_1.I80F48.fromU64(bank.feesWithdrawn));
        const feesUi = (0, utils_1.toUiDecimals)(fees, bank.mintDecimals);
        const feesInQuoteUi = feesUi * bank.uiPrice;
        if (feesInQuoteUi < 50) {
            return;
        }
        const ix = await client.program.methods
            .adminTokenWithdrawFees()
            .accounts({
            group: group.publicKey,
            admin: group.admin,
            bank: bank.publicKey,
            vault: bank.vault,
            tokenAccount,
        })
            .instruction();
        const tx = new web3_js_1.Transaction({ feePayer: wallet.publicKey }).add(ix);
        const simulated = await client.connection.simulateTransaction(tx);
        if (simulated.value.err) {
            console.log(`${bank.name} Sim error` + simulated.value.logs);
            throw simulated.value.logs;
        }
        instructions.push(ix);
    });
    const tokenOwnerRecordPk = await (0, spl_governance_1.getTokenOwnerRecordAddress)(constants_2.MANGO_GOVERNANCE_PROGRAM, constants_2.MANGO_REALM_PK, constants_2.MANGO_MINT, new web3_js_1.PublicKey(VSR_DELEGATE_FROM_PK));
    const [tokenOwnerRecord, proposals] = await Promise.all([
        (0, spl_governance_1.getTokenOwnerRecord)(client.connection, tokenOwnerRecordPk),
        (0, spl_governance_1.getAllProposals)(client.connection, constants_2.MANGO_GOVERNANCE_PROGRAM, constants_2.MANGO_REALM_PK),
    ]);
    const walletSigner = wallet;
    if (!DRY_RUN) {
        const proposalAddress = await (0, createProposal_1.createProposal)(client.connection, walletSigner, constants_2.MANGO_DAO_WALLET_GOVERNANCE, tokenOwnerRecord, PROPOSAL_TITLE ? PROPOSAL_TITLE : 'Withdraw token fees in mango-v4', PROPOSAL_LINK ?? '', Object.values(proposals).length, instructions, vsrClient, false);
        console.log(proposalAddress.toBase58());
    }
}
async function main() {
    try {
        await withdrawFeesToAdmin();
    }
    catch (error) {
        console.log(error);
    }
}
try {
    main();
}
catch (error) {
    console.log(error);
}
